#ifndef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#define INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H

// $Id: atrp.h,v 1.6 2001/10/04 12:27:36 ar Exp $

/* Copyright (C) 2001 ATRSoft GmbH 
   This file is part of the ATRSoft GmbH utilitys.

   The ATRSoft GmbH utilitys is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or any later version.

   The ATRSoft GmbH utilitys is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the ATRSoft GmbH utilitys; see the file LICENSE.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

// http://www.atrsoft.de
// mailto: info@atrsoft.de

// we use exceptions
#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

namespace ATRSoft {

namespace Utility {

union Chunk;

//------------------------------------------------------------------------------


class AtrpManager
{
    // manager class for the interface with memory management
    // and dll reference counting, dynamic linking and releasing
    // the base class, of cause. you have to derive your own manager
    // to do the basic stuff ....

public:
    
    // we have to use pointers to normal functions....
    typedef void (*Pfunc)();

public:
    
    AtrpManager
        (
	)
    {
	// we do not need to instantiate a manager.
	// it is only an interface at this point
	// but we deliver a default impl to start
	// so we have a ctor in here
    }

    virtual ~AtrpManager
        (
	);

    // this has to be defined: deliver properly aligned memory
    // of this size
    virtual Chunk*
    allocate
        (
	int	i_total_size_in_bytes
	) ;

    // this has to be defined: give the memory free
    virtual void
    deallocate
        (
	Chunk*	i_chunks	// address in memory
	) ;

    // this has to be defined: add a connection to the manager
    virtual void
    connect
        (
	) ;

    // this has to be defined: subtract a connection 
    // from the manager -- and free
    virtual void
    disconnect
        (
	) ;

    // this has to be defined: subtract a connection 
    // from the manager -- but NO free
    virtual void
    decrement
        (
	) ;

    // this has to be defined: deliver a factory which can be
    // casted to the needs of the implementation in the prg
    // factory. placeholder is a pointer of a void function...
    virtual Pfunc
    get_factory
        (
	) =0;
}; // end AtrpManager

union Chunk
{
    // memory tile -- see stroustop 2.nd edition
    // aligning requirements are fullfilled by it

    enum Memory_state
    {
	// state of a tile
	null     = 0,	// dirty trick - uninitialized 
	                // static mem is null per default ...
	good	 = 1,
	bad	 = 2,
	free	 = 3,
	deleted	 = 4
    };

    struct Info
    {
	// management infos for 1.st tile are here

	// callback object for free is here
	AtrpManager*  manager_;

	// length of memory -- should be a size_t ?? ...
	int	      length_;

	// count of connections for the object -- could be a short ??
	int	      count_;

	// state of object - see Memory_state, we save son bytes perhaps ...
	char	      state_;
    };

    struct Link
    {
	// callback object is here
	AtrpManager*  manager_;

	// length of this node
	int           length_;

	// simple linked list of free nodes
	Chunk*        next_;
    };

    // data -- can be
    int		        i_;	// an int ...
    char		c_;	// or a char ...
    void*		p_;	// or a pointer ...
    long double	        ld_;	// or a float
    
    Link		l_;	// or a node on a free-list
    
    Info		s_;     // or a real active object
}; // end Chunk

//------------------------------------------------------------------------

template < typename AB >
class Atrp
{
    // abstract base reference counted pointer -- interface seen only
    // needs help of factory classes for instantiation
    // knows only the interface class AB. the implementation is
    // not known at the interface side of the system
  
    // atrsoft smart pointer. 
    // benoetigt eine basisklasse. eine factory instanziert dann die 
    // eigentlichen instanzen.
    // die speicherverwaltung und die freigabe sind via manager geregelt.
    // der manager ist dabei ggf in einer sh lib oder einem dynmisch geladenen o file
    
    Chunk* chunk_;

protected:
    
    // konstruktor fuer die factory uebergabe. 
    // wird mit speicheradresse des genutzten speichers aufgerufen.
    explicit Atrp
        (
	void* i_p
	) throw ()
	:
        chunk_ (reinterpret_cast <Chunk*> (i_p))
    {
	// ctor for the prg factory -- uses void * because of c interface

	connect ();
    }
    
public:

    // copy konstruktor. 
    // zaehlt die referenz hoch und kopiert ansonsten nur den pointer
    Atrp
        (
	const Atrp& i_rhs
	) throw ()
	:
	chunk_ (i_rhs.chunk_)
    {
	connect ();
    }

    // destruktor.
    // entkoppelt sich vom verwalteten objekt. 
    // release erledigt den rest.
    // kein virtueller ctor da wir nicht von atrp ableiten ausser zu interface uebernahme
    // erweiterungen von atrp um member sind damit verboten.
    ~Atrp
        (
	) // throw whatever the dtor of AB (or derived classes) throws ...
    {
	// free of object is here - if needed
	// NO virtual ... we do not need a vtbl ...
  
	release ();
    }

    // zuweisung. 
    // einfach nur den aktuellen obj freigeben und dann wieder 
    // ankoppeln. ist das erforderlich wird altes objekt damit
    // freigegeben
    Atrp&
    operator=
        (
	const Atrp& i_rhs
	) // throw whatever the dtor of object throws ...
    {
	// IMPORTANT : compare the values of chunk pointers,
	// not the addresses of the ptr objects...

	if (chunk_ != i_rhs.chunk_)
	{
	    // first, free old object
	    
	    release ();

	    chunk_ = i_rhs.chunk_;

	    connect ();
	}
	
	return *this;
    }

    
private:

    // freigeben
    // macht mit dem manager des objekts die freigabe
    void
    release
        (
        )
    {
	// free object. decrement ref. counter, test 
	// state and destruct if needed

	// erst mal entkoppeln von dem objekt
	disconnect ();

	// nun der test : ist objekt nun waise
	if (get_count () != 0)
	    return;

	// nun test des zustand : darf ich loeschen ?
	if (get_state () == Chunk::good)
	{
	    // ok. gebe frei
	    destroy ();
	    return;
	}

	// nicht ok. nur status switchen 
	set_state_deleted ();

	// und manager freigeben
	free_manager ();
    }

public:

    // access functions

    // deliver the raw pointer
    // liefere den zeiger - nur in ausnahmefaellen erforderlich...
    AB*
    get_dump_object_pointer // long name is a psycho ... ;-)
        (
	) const throw ()
    {
	// instead of a conversion operator - you 
	// have to do this explicitly

	return reinterpret_cast <AB*> (chunk_);
    }

    // access to the object
    // zugriff auf das objekt...
    AB*
    operator->
        (
	) const throw ()
	{
	    return reinterpret_cast <AB*> (chunk_);
	}

    AB&
    operator*
        (
        ) const throw ()
    {
	return *reinterpret_cast <AB*> (chunk_);
    }

    // the count of active maintaining referencers
    // anzahl verantwortliche
    int
    get_count
        (
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;
  
	return mem->s_.count_;
    }

    // set the state
    // status setzen
    void
    set_state
        (
	int i_s
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;
  
	mem->s_.state_ = char (i_s);
    }

    void
    set_state_free
        (
	) const throw ()
    {
	set_state (Chunk::free);
    }

    void
    set_state_good
        (
	) const throw ()
    {
	set_state (Chunk::good);
    }

    void
    set_state_deleted
        (
	) const throw ()
    {
	set_state (Chunk::deleted);
    }

    void
    set_state_bad
        (
	) const throw ()
    {
	set_state (Chunk::bad);
    }

    Chunk::Memory_state
    get_state
        (
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;

	return Chunk::Memory_state (mem->s_.state_);
    }

    // compare with pointer identity
    // vergleich mittels identitaet der verwalteten zeiger
    bool
    operator ==
        (
	const Atrp<AB>& i_rhs
	) const throw ()
    {
	// same objects if address is same
  
	return chunk_ == i_rhs.chunk_;
    }

    // we need this - if it is ok ?
    // we think in terms of contignoous memory here ...
    // wups. das ist zwar notwendig, aber ob es richtig ist ?
    // annahme von kontinuierlichem speicher
    bool
    operator <
        (
	const Atrp<AB>& i_rhs
	) const throw ()
    {
	// stl container needs this for ordering 
	// key values -- see map and set
	// assume contignous memory model here
	
	return (reinterpret_cast <char*> (chunk_)
		- reinterpret_cast <char*> (i_rhs.chunk_)) < 0;
    }

private:

    // forbidden -- use Atrp_d instead
    Atrp
        (
	);

    // ok. freigeben - und zwar speicher
    void
    destroy
        (
	)
    {
	// IMPORTANT: this needs a virtual destructor for AB if AB is a
	// base class - otherwise this will not work correctly
	
	set_state_deleted ();
		
	Chunk* const mem = chunk_ - 1;

	// den manager merken...
	AtrpManager *manager = mem->s_.manager_;

	try
	{
	    AB* obj = reinterpret_cast <AB*> (chunk_);
      
	    // please, dont throw any exception here - PLEASE !!
      
	    obj->~AB ();		// BC 5.02 problem : obj->~AB (); doesnt work

	    free_memory ();

	    manager->disconnect ();

	    return;
	}

	catch (...)
	{
	    free_memory ();	// no leaks
      
	    manager->disconnect ();

	    throw; // BC 5.02 problem rethrow stack overflow ...
	}
    }

    // speicher freigeben
    void
    free_memory
        (
	) throw ()
    {
	// free the memory via manager

	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->deallocate (mem);
    }

    // manager referenzen freigeben
    void
    free_manager
        (
	) throw ()
    {
	// free the manager via manager
  
	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->disconnect ();
    }

    // objekt ankoppeln
    void
    connect
        (
	) const throw ()
    {
	// register a new connection to the memory

	Chunk* const mem = chunk_ - 1;

	++ (mem->s_.count_);
    }

    // objekt abkoppeln
    void
    disconnect
        (
	) const throw ()
    {
	// lose a connection

	Chunk* const mem = chunk_ - 1;

	-- (mem->s_.count_);
    }
    
}; // end Atrp


//------------------------------------------------------------------------

template < typename AB >
class Atrp_d
{
    // abstract base reference counted pointer -- interface seen only
    // needs help of factory classes for instantiation
    // knows only the interface class AB. the implementation is
    // not known at the interface side of the system

    // has default ctor because of brain damaged libs and comps...

    // das ist ein hack fuer libs und comps die einen default ctor brauchen.
    // eigentlich ist der vom konzept her nicht notwendig...
    
    // der verwaltete objektspeicher
    Chunk* chunk_;

protected:

    explicit Atrp_d
        (
	void* i_p
	) throw (std::runtime_error)
	:
	chunk_ (reinterpret_cast <Chunk*> (i_p))
    {
	// ctor for the prg factory -- use void * because of c interface

	// well - we dont allow this kind of shit ...
	if (chunk_ != 0)
	{
	    connect ();
	    return;
	}
	 
	throw std::runtime_error("illegal initialization of Atrp_d : void* is NULL");
    }

public:

    // defualt ctor
    Atrp_d
        (
	) throw ()
	:
	chunk_ (0)
    {
	// some libs and comps need this kind of stuff ...
	// for example take code warrior 4 and std::list<> ...
    }

    Atrp_d
        (
	const Atrp_d& i_rhs
	) throw ()
	:
	chunk_ (i_rhs.chunk_)
    {
	if (chunk_)
	{
	    connect ();
	}
    }

    Atrp_d
        (
	const Atrp <AB>& i_rhs
	) throw ()
	:
	chunk_ (reinterpret_cast <Chunk*> (i_rhs.get_dump_object_pointer ()))
    {
	// this direction is OK : assign / convert a Atrp to a Atrp_d
	// WARNING: the other direction is NOT OK -- dont do it ...

	connect ();
    }

    ~Atrp_d
        (
	) // throw whatever the dtor of AB (or derived classes) throws ...
    {
	// free of object is here - if needed
	// NO virtual ... we do not need a vtbl ...

	if (chunk_)
	{
	    release ();
	}
    }

    Atrp_d&
    operator=
        (
	const Atrp_d& i_rhs
	) // throw whatever the dtor of object throws ...
    {
	// IMPORTANT : compare the values of chunks, not 
	// the addresses of selfs ...

	if (chunk_ != i_rhs.chunk_)
	{
	    // first, free old object
	    if (chunk_)
	    {
		release ();
	    }

	    // now, copy address
	    chunk_ = i_rhs.chunk_;

	    // and correct count
	    if (chunk_)
	    {
		connect ();
	    }
	}

	return *this;
    }


private:

    void
    release
        (
	) // throw what AB dtor can throw ...
   {
       // free object. decrement ref. counter, test state 
       // and destruct if needed

       disconnect ();

       if (get_count () != 0)
	   return;

       if (get_state () == Chunk::good)
       {
	   destroy ();
	   return;
       }
       
       // nicht ok. nur status switchen 
       set_state_deleted ();

       free_manager ();
   }

public:

    // access functions

    AB*
    get_dump_object_pointer // long name is a psycho ... ;-)
        (
	) const throw ()
    {
	// instead of a conversion operator - you have 
	// to do this explicitly
	return reinterpret_cast <AB*> (chunk_);
    }

    AB*
    operator->
        (
	) const throw ()
    {
	return reinterpret_cast <AB*> (chunk_);
    }

    AB&
    operator*
        (
	) const throw ()
    {
	return *reinterpret_cast <AB*> (chunk_);
    }

    int
    get_count
        (
	) const throw ()
    {
	if (chunk_ == 0)
	{
	    return -1;
	}
	
	Chunk* const mem = chunk_ - 1;
	
	return mem->s_.count_;
    }

    void
    set_state
        (
	int i_s
	) const throw ()
    {
	if (chunk_ == 0)
	{
	    return ;
	}
	
	Chunk* const mem = chunk_ - 1;
	
	mem->s_.state_ = char (i_s);
    }

    void
    set_state_free
        (
	) const throw ()
    {
	set_state (Chunk::free);
    }

    void
    set_state_good
        (
	) const throw ()
    {
	set_state (Chunk::good);
    }

    void
    set_state_deleted
        (
	) const throw ()
    {
	set_state (Chunk::deleted);
    }

    void
    set_state_bad
        (
	) const throw ()
    {
	set_state (Chunk::bad);
    }

    Chunk::Memory_state
    get_state
        (
	) const throw ()
    {
	if (chunk_ == 0)
	{
	    return Chunk::null;
	}

	Chunk* const mem = chunk_ - 1;

	return Chunk::Memory_state (mem->s_.state_);
    }

    bool
    operator ==
        (
	const Atrp_d <AB>& i_rhs
	) const throw ()
    {
	return chunk_ == i_rhs.chunk_;
    }

    bool
    operator <
        (
	const Atrp_d <AB>& i_rhs
	) const throw ()
    {
	// well. this is needed for ordered containers with a abcptr_d as key ...
	// for example a set or map ...

	return (reinterpret_cast <char*> (chunk_)
		- reinterpret_cast <char*> (i_rhs.chunk_)) < 0;
    }

private:

    void
    destroy
        (
	)
    {
	// IMPORTANT: this needs a virtual destructor for AB if AB is a
	// base class - otherwise this will not work correctly ...
  
	Chunk* const mem = chunk_ - 1;
	
	// we need this later after deletion of tile 0 - so save it now
	AtrpManager *manager = mem->s_.manager_;

	try
	{
	    AB* obj = reinterpret_cast <AB*> (chunk_);

	    // please, dont throw an exception here - PLEASE !!

	    obj->~AB ();		// BC 5.02 problem : obj->~AB (); doesnt work
								// use obj->AB::~AB() instead...
	    free_memory ();

	    manager->disconnect ();
	    
	    return;
    	}

	catch (...)
	{
	    // but if you do, we are ready for it ...
	    free_memory ();	// no leaks

	    manager->disconnect ();
	    
	    throw;
	}
    }

    void
    free_memory
        (
	) throw ()
    {
	// free the memory via manager

	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->deallocate (mem);
    }
    
    void
    free_manager
        (
	) throw ()
    {
	// free the manager via manager

	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->disconnect ();
    }

    void
    connect
        (
	) const throw ()
    {
	// register a new connection to the memory

	Chunk* const mem = chunk_ - 1;

	++ (mem->s_.count_);
    }
    
    void
    disconnect
        (
	) const throw ()
    {
	// lose a connection

	Chunk* const mem = chunk_ - 1;

	-- (mem->s_.count_);
    }

}; // end Atrp_d


} // end Utility

} // end Atrsoft

//----------------------------------------------------------------------

// poor man solution for inplace construction of a abcptr
// and usage of any ctor of the derived class
// example: ATRP(X, ptr, Y, (parm1,parm2,parm3));
// where Y is a derived class from class X
// and has a ctor of the form Y (T1 parm1, T2 parm2, T3 parm3)
// because you have to know the ctor you have to include
// interface and implementation together
// you have also to include abcptrinit and abcptrdllinit
// for the helper classes - its really a poor solution...

#ifndef ATRP

//#include <ar/utility/abcptrinit.h>
//#include <ar/utility/abcptrdllinit.h>

// use a manager
#define ATRPMANAGER(__classX,__namePtr,__classY,__params,__manager) \
	ATRSoft::Utility::Atrp < __classX > __namePtr ( \
   	ATRSoft::Utility::AtrpInit < __classX > \
      	( \
         	new (ATRSoft::Utility::AtrpDLLInit ( \
		    		((sizeof(__classY) + sizeof(ATRSoft::Utility::Chunk) - 1) \
			/ sizeof (ATRSoft::Utility::Chunk) + 1) \
			* sizeof(ATRSoft::Utility::Chunk), \
				__manager).get_adr ()) __classY __params \
         ) \
      )

// use the default manager
#define ATRP(__classX,__namePtr,__classY,__params) \
	ATRPMANAGER(__classX,__namePtr,__classY,__params,  \
   		ATRSoft::Utility::g_default_manager_p)

#endif

#endif

// end of file

